from setuptools import setup

setup(
	name="oscillations",
	version="1.1.1",
	description="Calculate neutrino oscillation probabilities easily",
	long_description="Calculate neutrino oscillation probabilities easily",
	url="https://github.com/discully/oscillations",
	author="Daniel I. Scully",
	author_email="discully@users.noreply.github.com",
	packages=["oscillations"],
	install_requires=[],
	classifiers=["Programming Language :: Python :: 3"]
)
